
// 
// SuperPlayer Bluetooth (BLE) Controller 
// Build from different sources (www)
// JesperLykke [User :: Lykkedk @ diyaudio.com]
//

// - Sleep stuff -

unsigned long startMillis;
unsigned long currentMillis;

// period ::
//1 Hour  3,600,000 ms
//2 Hours 7,200,000 ms
//3 Hours 10,800,000 ms
//4 Hours 14,400,000 ms
const unsigned long period = 3600000;  //the value is a number of milliseconds

unsigned long BLE_startMillis;
unsigned long BLE_currentMillis;
const unsigned long BLE_period = 3000;  //Adjust a bit to fit the wakeup/vs. when first BLE send is fired

// - Encoder stuff -
#include <ESP32Encoder.h>
#include <BfButton.h>

// - BLE stuff -

#include <BLEDevice.h>
#include <BLEServer.h>
#include <BLEUtils.h>
#include <BLE2902.h>

// - Sleep stuff -

#define BUTTON_PIN_BITMASK 0x200000000 // 2^33 in hex
RTC_DATA_ATTR int bootCount = 0;

// - Encoder -

const unsigned int btnPin = 25;
BfButton btn(BfButton::STANDALONE_DIGITAL, btnPin, false, HIGH); // I added a pull-down myself
ESP32Encoder encoder;

// Encoder values for Volume control
int32_t Volume = 0;
int32_t OLD_Volume = 0;
uint32_t BLE_encoder = 0;
uint32_t OLD_BLE_encoder = 0;

// - BLE stuff -

BLEServer* pServer = NULL;
BLECharacteristic* pCharacteristic = NULL;
bool deviceConnected = false;
bool oldDeviceConnected = false;

#define SERVICE_UUID        "4fafc201-1fb5-459e-8fcc-c5c9c331914b"
#define CHARACTERISTIC_UUID "beb5483e-36e1-4688-b7f5-ea07361b26a8"

class MyServerCallbacks: public BLEServerCallbacks {
    void onConnect(BLEServer* pServer) {
      deviceConnected = true;
    };

    void onDisconnect(BLEServer* pServer) {
      deviceConnected = false;
    }
};

// - Handle Sleep -

void print_wakeup_reason(){
  esp_sleep_wakeup_cause_t wakeup_reason;

  wakeup_reason = esp_sleep_get_wakeup_cause();

  switch(wakeup_reason)
  {
    case ESP_SLEEP_WAKEUP_EXT0 : Serial.println("Wakeup caused by external signal using RTC_IO"); break;
    case ESP_SLEEP_WAKEUP_EXT1 : Serial.println("Wakeup caused by external signal using RTC_CNTL"); break;
    case ESP_SLEEP_WAKEUP_TIMER : Serial.println("Wakeup caused by timer"); break;
    case ESP_SLEEP_WAKEUP_TOUCHPAD : Serial.println("Wakeup caused by touchpad"); break;
    case ESP_SLEEP_WAKEUP_ULP : Serial.println("Wakeup caused by ULP program"); break;
    default : Serial.printf("Wakeup was not caused by deep sleep: %d\n",wakeup_reason); break;
  }
}

// - Handle BLE -

void Ble() { // Start void BLE


    // Only fire BLE stuff, after a short time when awake from deep sleep
    // BLE_period is that time configured first in this code 
    BLE_currentMillis = millis();
    if (BLE_currentMillis - BLE_startMillis >= BLE_period) { // Start BLE first fire 
           
    
    if (BLE_encoder!=OLD_BLE_encoder) {

    Serial.println("BLE_encoder value :");
    Serial.println(BLE_encoder);

    startMillis = millis();  // Reset initial gotosleep start time
        
    if (deviceConnected) {     
    pCharacteristic->setValue(BLE_encoder);
    // pCharacteristic->setValue(value);
        
    pCharacteristic->notify();
    delay(5); // bluetooth stack will go into congestion, if too many packets are sent, in 6 hours test i was able to go as low as 3ms
    delay(500);
    }

    // startMillis = millis();  // Reset initial gotosleep start time 
    OLD_BLE_encoder = BLE_encoder;
    }
    
    // disconnecting
    if (!deviceConnected && oldDeviceConnected) {
        delay(500); // give the bluetooth stack the chance to get things ready
        pServer->startAdvertising(); // restart advertising
        Serial.println("start advertising");
        Serial.println("DisConnecting");
        oldDeviceConnected = deviceConnected;
     }
    // connecting
    if (deviceConnected && !oldDeviceConnected) {
        Serial.println("Connecting");
        // do stuff here on connecting
        oldDeviceConnected = deviceConnected;
    }
    } // End BLE first fire
} // End void BLE

// - Handle Encoder Button -

void pressHandler (BfButton *btn, BfButton::press_pattern_t pattern) {
  Serial.print(btn->getID());
  switch (pattern) {
    case BfButton::SINGLE_PRESS: // Does not work very well!
      Serial.println(" pressed.");
      BLE_encoder = 75000; 
      break;
    case BfButton::DOUBLE_PRESS: // Works fine
      Serial.println(" double pressed.");
      BLE_encoder = 50000;
      break;
    case BfButton::LONG_PRESS:  // Works fine
      Serial.println(" long pressed.");
      BLE_encoder = 25000;
      break;
  }
}

// - Handle Encoder data -

void ReadEncoder () {
  // Loop and read the count
  Volume = encoder.getCount();
  if (Volume!=OLD_Volume) {
    Serial.println(Volume);
    OLD_Volume = Volume; 
    
  }

  BLE_encoder = Volume;
      
}

// - SETUP -

void setup(){
	
	Serial.begin(115200);

  startMillis = millis();  // Initial gotosleep start time
  BLE_startMillis = millis();  // First BLE fire start counter
  
  //Increment boot number and print it every reboot - Sleep
  ++bootCount;
  Serial.println("Boot number: " + String(bootCount));

  //Print the wakeup reason for ESP32
  print_wakeup_reason();

  esp_sleep_enable_ext0_wakeup(GPIO_NUM_33,1); //1 = High, 0 = Low

// - BLE stuff -

      // Create the BLE Device
    BLEDevice::init("ESP32");

    // Create the BLE Server
    pServer = BLEDevice::createServer();
    pServer->setCallbacks(new MyServerCallbacks());

    // Create the BLE Service
    BLEService *pService = pServer->createService(SERVICE_UUID);

    // Create a BLE Characteristic
    pCharacteristic = pService->createCharacteristic(
                      CHARACTERISTIC_UUID,
                      BLECharacteristic::PROPERTY_READ   |
                      BLECharacteristic::PROPERTY_WRITE  |
                      BLECharacteristic::PROPERTY_NOTIFY |
                      BLECharacteristic::PROPERTY_INDICATE
                    );

    // https://www.bluetooth.com/specifications/gatt/viewer?attributeXmlFile=org.bluetooth.descriptor.gatt.client_characteristic_configuration.xml
    // Create a BLE Descriptor
    pCharacteristic->addDescriptor(new BLE2902());

    // Start the service
    pService->start();

    // Start advertising
    BLEAdvertising *pAdvertising = BLEDevice::getAdvertising();
    pAdvertising->addServiceUUID(SERVICE_UUID);
    pAdvertising->setScanResponse(false);
    pAdvertising->setMinPreferred(0x0);  // set value to 0x00 to not advertise this parameter
    BLEDevice::startAdvertising();
    Serial.println("Waiting a client connection to notify...");

// - Button stuff -
	
	btn.onPress(pressHandler)
     .onDoublePress(pressHandler) // default timeout
     .onPressFor(pressHandler, 1000); // custom timeout for 1 second

// - Encoder stuff -
  
	// Enable the weak pull down resistors
	//ESP32Encoder::useInternalWeakPullResistors=DOWN;
	// Enable the weak pull up resistors
	ESP32Encoder::useInternalWeakPullResistors=UP;
  //CLK (A pin) - to any microcontroler intput pin with interrupt -> in this example pin 32
  //DT (B pin) - to any microcontroler intput pin with interrupt -> in this example pin 21
  //SW (button pin) - to any microcontroler intput pin -> in this example pin 25
  //VCC - to microcontroler VCC (then set ROTARY_ENCODER_VCC_PIN -1) or in this example pin 25
  //GND - to microcontroler GND
	encoder.attachHalfQuad(32, 21);
  //encoder2.attachHalfQuad(32, 21);
		
	// set starting count value after attaching
	encoder.setCount(5000);

	// clear the encoder's raw count and set the tracked count to zero
	//encoder2.clearCount();
	//Serial.println("Encoder Start = "+String((int32_t)encoder.getCount()));
	// set the lastToggle
	//encoder2lastToggled = millis();


}

// - Loop section -

void loop(){

  ReadEncoder();
  btn.read();
  Ble();

  currentMillis = millis(); // Grab the current time
  if (currentMillis - startMillis >= period) { // Check if period before gotosleep is gone (period = time before we gotosleep)
  Serial.println("Going to sleep now");
  esp_deep_sleep_start();
  
  }
  
}
